; vim: sts=4 sw=8 et ai fdm=marker syntax=avrasm

.nolist
.include "../m8def.inc"
.list
.include "../fifo_handling.inc"

;   flags and aliases
; =============================================================================

; serial port
.equ CLOCK = 16000000
.equ BAUD = 19200
.equ UBRRVAL = CLOCK/(BAUD*16)-1

; fifo sizes
.equ URSize = 32         ; uart receive fifo size
.equ UTSize = 128        ; uart transmit fifo size

; leds
.equ UARTLED = 5

; flags - receiving states 
.equ    FlagPinState        = 0     ; stores the last pin value (first half of the bit)
.equ    FlagCounterStarted  = 1     ; counter is running, and compare interrupt is active
.equ    FlagFirstBitDone    = 2     ; an initial bit has been received, we still need to do init stuff
.equ    FlagReceiving       = 3     ; receive init is done. i.e. we are receiving the real bits
.equ    FlagFirstHalfDone   = 4     ; first or second part in bit ? (ManchesterCode)
                                    

;   fifos {{{
; =============================================================================
.dseg
        ;   fifo: UReceiveFifo
        ; ==========================
        UReceiveFifo:
            .byte   FifoControlSize + URSize

        ;   fifo: UTransmitFifo
        ; ==========================
        UTransmitFifo:
            .byte   FifoControlSize + UTSize

.cseg
; }}}


;   register definitions
; =============================================================================
.def    l1      =   r8      ; low working registers
.def    l2      =   r9
.def    w1      =   r16     ; working registers
.def    w2      =   r17
.def    w3      =   r18
.def    w4      =   r19
.def    Flags   =   r20     ; global flags

.def    bitcount =  r21      
.def    ir_command = r22        ; received ir command
.def    ir_address = r23        ; received ir command
.def    half_bit_timeL = r24    ; time measured for the first half bit  
.def    half_bit_timeH = r25    ; - used to synchronize the receive process


;   interrupt table {{{
; =============================================================================
.ORG 0x0                            ; reset vector
        rjmp reset
.ORG INT0addr                       ; external interrupt 0 handler
        rjmp INT0Handler
.ORG OVF1addr                       ; timer1 overflow interrupt handler
        rjmp OVF1Handler
.ORG OC1Aaddr                       ; compare interrupt handler 
        rjmp OC1AHandler
.ORG URXCaddr                       ; USART receive complete interrupt handler
        rjmp URXCHandler
.ORG UDREaddr                       ; USART data register empty handler
        rjmp UDREHandler
.ORG 0x13                           ; program start

; }}}


;   SERIALDEBUG output routines {{{
; =============================================================================

;   DEBUG_write_utf {{{
; =============================================================================
; saves registers and calls write_uart_transmit_fifo
DEBUG_write_utf:
        push w2
        push l1
        push l2
        push YH
        push YL
        rcall write_uart_transmit_fifo
        pop YL
        pop YH
        pop l2
        pop l1
        pop w2
        ret
; }}}

; writes an immediate value
.macro DEBUG_WRITE_I
.ifdef SERIALDEBUG
    push w1
    ldi w1, @0
    rcall DEBUG_write_utf
    pop w1
.endif
.endmacro

; writes an register value
.macro DEBUG_WRITE_R
.ifdef SERIALDEBUG
    push w1
    mov w1, @0
    rcall DEBUG_write_utf
    pop w1
.endif
.endmacro

; }}}


;   DEVELBOARD output routines {{{
; =============================================================================

; toggles a led
.macro DEBUG_LED_TOGGLE
.ifdef DEVELBOARD
    push w1
    push w2
    in w1, PORTD
    ldi w2, (1<<@0)<<2
    eor w1, w2
    out PORTD, w1
    pop w2
    pop w1
.endif
.endmacro

; switch on a led
.macro DEBUG_LED_ON
.ifdef DEVELBOARD
    cbi PORTD, @0+2
.endif
.endmacro

; switch off a led
.macro DEBUG_LED_OFF
.ifdef DEVELBOARD
    sbi PORTD, @0+2
.endif
.endmacro


; }}}


;   function: reset interrupt handler {{{
; =============================================================================
reset:

        ;  init stack
        ; ========================================
        ldi w1, low(RAMEND)
        out SPL, w1
        ldi w1, high(RAMEND)
        out SPH, w1

        ;  init uart
        ; ========================================
        ldi w1, high(UBRRVAL)
        out UBRRH, w1
        ldi w1, low(UBRRVAL)
        out UBRRL, w1
        ldi w1, (1<<URSEL)|(3<<UCSZ0) ; frame format: 8N1
        out UCSRC, w1

        ;  init fifos
        ; ========================================
        MACRO_FIFO_INIT UReceiveFifo
        MACRO_FIFO_INIT UTransmitFifo

        ;  enable uart interrupt
        ; ========================================
        ldi w1, (1<<RXCIE)|(1<<RXEN)|(1<<TXEN); receiver interrupt complete enable, receiver enable,
                                        ; transmitter enable
        out UCSRB, w1

        ;  init led output
        ; ========================================
        ldi w1, 0b111                   ; use PB0-PB2 as output
        out DDRB, w1

        ldi w1, 0b111                   ; all leds off
        out PORTB, w1

        .ifdef DEVELBOARD
            ldi w1, (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) ; configure PD3-PD7 as output, PD2 as input (int0)
            out DDRD, w1
            ldi w1, 0xff                    ; all channels high, leds off and pull-up active
            out PORTD, w1
        .endif

        ;  init timer1
        ; ========================================
        clr w1                      ; no clock, normal mode (timer stopped)
        out TCCR1B, w1
        clr w1                      ; nothing else
        out TCCR1A, w1
        ;ldi w1, (1<<TOIE1)          ; enable overflow interrupt
        ;out TIMSK, w1
        clr w1                      ; reset counter
        out TCNT1H, w1
        out TCNT1L, w1

        ;  init external interrupt
        ; ========================================

        ldi w1, (1<<ISC00)              ; configure int0 so that any logical change will generate an
                                        ; interrupt, (ISC01, ISC00) == (0, 1)
        out MCUCR, w1
        ldi w1, (1<<INT0)               ; configure int0 interrupt request enable
        out GICR, w1

        ; init receiving flags 
        ; ========================================

        clr Flags

        ;  enable interrupts
        ; ========================================
        sei

        ;  output system bootup message to uart
        ; ========================================
        rcall output_string
        .db "bc", 0, 0

        ;  start main procedure
        ; ========================================
        rjmp main

; }}}


;   function: int0 interrupt handler {{{
; =============================================================================
INT0Handler:
        push w1                     ; save w1
        in w1, SREG                 ; save the status register
        push w1

        sbrc Flags, FlagCounterStarted        ; if not CounterStarted -> init the counter
        rjmp INT0Handler_calc_half_bit_time   ; otherwise jump to second init step

    INT0Handler_init_counter:               
                                                            ;    (- 1 -)
        ; initialize the timer                       
        clr w1                      ; reset counter
        out TCNT1L, w1
        out TCNT1H, w1

        ldi w1, (1<<CS11) | (1<<CS10)    ; prescaler 64
        out TCCR1B, w1
        clr w1                           ; only set the prescaler nothing else
        out TCCR1A, w1

        clr half_bit_timeL                   ; clear possibly recorded half-bit times
        clr half_bit_timeH
        clr bitcount                         ; no bits received
        clr ir_command

        ldi Flags, (1<<FlagCounterStarted)   ; set timer started flag

        .ifdef SERIALDEBUG  
                ldi w1, 'A'                  ; first flank noticed  * A * 
                rcall DEBUG_write_utf
        .endif

        rjmp INT0Handler_cleanup

    INT0Handler_calc_half_bit_time:                  
                                                            ;    (- 2 -)

        sbrc Flags, FlagFirstBitDone        ; if first bits were received JUMP to init actual receiving
        rjmp INT0Handler_rc5_init_done

        in half_bit_timeL, TCNT1L           ; load counter
        in half_bit_timeH, TCNT1H

        .ifdef SERIALDEBUG                  ; debug output      
                ldi w1, 'T'                 
                rcall DEBUG_write_utf

                mov w1, half_bit_timeL
                rcall DEBUG_write_utf
                mov w1, half_bit_timeH
                rcall DEBUG_write_utf
        .endif

        ldi Flags, (1<<FlagFirstBitDone) | (1<<FlagCounterStarted); make sure we proceed to the next stage

        rjmp INT0Handler_cleanup

    INT0Handler_rc5_init_done:

                                                            ;    (- 3 -)
        ; enable compare timer interrupt

        ; FIXME..should be true 16bit calc stuff..
        ;mov w1, half_bit_timeL
        ;lsr w1
        ;add half0_bit_timeL, w1
        ldi w1, 25
        add half_bit_timeL, w1
        out OCR1AL, half_bit_timeL              
        out OCR1AH, half_bit_timeH

        ldi w1, (1<<OCIE1A)
        out TIMSK, w1
        
        ; disable INT0 interrupt  (we dont want receiving to be disturbed!)
        in w1, GICR
        cbr w1, (1<<INT0)
        out GICR, w1

; ??
;        clr Flags, (1<<FlagFirstBitDone)        ; start out with the next first bit
; ??

        .ifdef SERIALDEBUG                  ; starting to receive the useful stuff  * B * 
                ldi w1, 'B'
                rcall DEBUG_write_utf
        .endif


    INT0Handler_cleanup:


        ; the actual cleanup part
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti
; }}}


;   function: timer1 overflow interrupt handler {{{
; =============================================================================
OVF1Handler:
        push w1                     ; save w1
        in w1, SREG                 ; save the status register
        push w1

        clr Flags

        ; disable the timer compare interrupt, and reset counter
        clr w1                     ; stop the timer
        out TCCR1A, w1
        out TCNT1L, w1
        out TCNT1H, w1

        clr half_bit_timeL
        clr half_bit_timeH

        ; reactivate the int0 interrupt 
        in w1, GICR
        sbr  w1, (1<<INT0)
        out GICR, w1

        .ifdef SERIALDEBUG
                ldi w1, 'O'
                rcall DEBUG_write_utf
        .endif

        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti
; }}}


;   function: timer1 compare interrupt handler {{{
; =============================================================================
OC1AHandler:
        push w1                     ; save w1
        in w1, SREG                 ; save the status register
        push w1
        push w2
        push w3

        clr w1
        out TCNT1L, w1
        out TCNT1H, w1
        

        sbrs Flags, FlagFirstHalfDone       ; which part of the bit are we at?
        rjmp OCI1AHandler_first_half_bit    ; first part -> just set FirstHalfFlag and store state


        cbr Flags, (1<<FlagFirstHalfDone)   ; we received a second part of a bit, get ready for next first

        inc bitcount
        lsl ir_command              ; FIXME, for now we only use ir_command later we also want ir_address 
        in w1, PIND
        ; check to see if there actually was a change... low/high .. or high/low ... 
        ; otherwise there is no bit and we need to exit 
        ; FIXME ... this is the and and eor stuff i couldnt do  

       
        sbrs w1, PIND2              ; if bit is set, skip the jump
        rjmp OCI1AHandler_bit_zero
        sbr ir_command, 1           ; 1 -> bitmask 0x00000001
        .ifdef IR_DEBUG
                ldi w1, '-'
                rcall DEBUG_write_utf
        .endif
        rjmp OCI1AHandler_check_if_done

OCI1AHandler_bit_zero:
        .ifdef IR_DEBUG
                ldi w1, '.'
                rcall DEBUG_write_utf
        .endif
        cbr ir_command, 1

OCI1AHandler_check_if_done:


        cpi bitcount, 12
        brne OCI1AHandler_cleanup

        ; if we got here, we received a full command 

        ldi w1, 0b110                   ; red leds on... indicating that the rc5 init stuff
                                        ; up to here worked
        out PORTB, w1

   ;     .ifdef SERIALDEBUG
                ldi w1, 'F'                 ; successfully finished a 12 bit sequence
                rcall DEBUG_write_utf

                mov w1, ir_command
                rcall DEBUG_write_utf
   ;     .endif

        clr Flags
        clr bitcount

        ; enable the INT0 interrupt for the next receiving
        in w1, GICR
        sbr w1, (1<<INT0)
        out GICR, w1

        ; disable the timer compare interrupt, and reset counter
        ldi w1, 0                   ; stop the timer
        out TCCR1B, w1
        clr w1
        out TCNT1L, w1
        out TCNT1H, w1

        rjmp OCI1AHandler_cleanup


OCI1AHandler_first_half_bit:
        sbr Flags, (1<<FlagFirstHalfDone)

        ; check the current state of the pin 
        in w1, PIND
        sbrs w1, PIND2              ; if bit is set, skip the jump
        rjmp OCI1AHandler_first_half_zero
        sbr Flags, (1<<FlagPinState)
        
        rjmp OCI1AHandler_cleanup
        
        OCI1AHandler_first_half_zero:

        cbr Flags, (1<<FlagPinState)


OCI1AHandler_cleanup:
        pop w3
        pop w2
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti
    
; }}}


;   function: uart receive complete interrupt handler {{{
; =============================================================================
URXCHandler:
        push w1                     ; save w1
        in w1, SREG                 ; save the status register
        push w1

        MACRO_FIFO_LOAD_COUNTER UReceiveFifo, w1 ; load counter
        cpi w1, URSize              ; compare counter and fifo size
        breq URXDiscardData         ; if counter == fifosize, just discard the byte received  1 false/2 true

        push w2                     ; save some more registers
        push YL                     ; save Y
        push YH

        DEBUG_LED_TOGGLE UARTLED    ; announce uart traffic by toggling led

        in w1, UDR                  ; read data byte

                                    ; store data into receive fifo
        MACRO_FIFO_STORE UReceiveFifo, w1, w2, URSize

URXCleanup:
        pop YH                      ; restore Y
        pop YL
        pop w2
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti

URXDiscardData:
        in w1, UDR                  ; clear interrupt by reading the received byte
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti

; }}}

;   function: uart data register empty handler {{{
; =============================================================================
UDREHandler:
        push w1                     ; save w1
        in w1, SREG                 ; save the status register
        push w1

        MACRO_FIFO_LOAD_COUNTER UTransmitFifo, w1   ; load uart transmit fifo counter into w1
        cpi w1, 1
        brsh UDRETransmitByte       ; if w1 >= 1, jump away
                                    ; else:
        in w1, UCSRB                ; disable data register empty interrupt
        cbr w1, (1<<UDRIE)
        out UCSRB, w1

        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti

UDRETransmitByte:
        push w2                     ; save registers
        push w3
        push YL                     ; save Y
        push YH

        MACRO_FIFO_READ UTransmitFifo, w1, UTSize
        out UDR, w1                 ; write data to uart data register

        pop YH                      ; restore Y
        pop YL
        pop w3                      ; restore registers
        pop w2
        pop w1                      ; restore the status register
        out SREG, w1
        pop w1                      ; restore w1
        reti
; }}}



;   function: main
; =============================================================================
main:

        rjmp main



;   function: write_uart_transmit_fifo {{{
; =============================================================================
; data byte is loaded into w1
write_uart_transmit_fifo:
        MACRO_FIFO_LOAD_COUNTER UTransmitFifo, w2   ; load uart transmit fifo counter into w2
        cpi w2, UTSize ; compare counter and fifo size
        breq write_uart_discard_data ; if counter == fifosize, just discard the byte received

        movw l2:l1, YH:YL           ; save Y

        MACRO_FIFO_STORE UTransmitFifo, w1, w2, UTSize
        sbi UCSRB, UDRIE            ; enable data register empty interrupt
        movw YH:YL, l2:l1           ; restore Y

write_uart_discard_data:

        ret
; }}}

;   function: output_string {{{
; =============================================================================
output_string:
        pop ZH              ; load string address from the stack
        pop ZL
        lsl ZL              ; shift left Z (using carry)
        rol ZH
output_next_char:
        lpm w1, Z+          ; load next byte into w1 and increment Z
        cpi w1, 0           ; if it's not the end of the string
        breq output_string_end ; jump away

        rcall write_uart_transmit_fifo ; transmit char

        rjmp output_next_char ; loop

output_string_end:
        adiw ZL, 1          ; increment Z to point to the next instruction after the string
        lsr ZH              ; shift right Z (using carry)
        ror ZL
        push ZL             ; push Z back onto the stack
        push ZH
        ret

; }}}
